#include "RigidBody.h"

#include <cmath>

RigidBody::RigidBody(StateList* sl) :
    m_Pos(sl), m_Vel(sl), m_BodyRate(sl), m_Quat(sl)
{
}

void RigidBody::Initialize(Vector<3> pos, Vector<3> vel, Vector<3> euler,
    Vector<3> body_rate)
{
    m_Pos.ic = pos;
    m_Vel.ic = vel;
    m_BodyRate.ic = body_rate;
    m_Euler = euler;

    // Set up the direction cosine matrix
    double sa = sin(m_Euler[0]);
    double ca = cos(m_Euler[0]);

    double sb = sin(m_Euler[1]);
    double cb = cos(m_Euler[1]);

    double sc = sin(m_Euler[2]);
    double cc = cos(m_Euler[2]);
 
    // Direction cosine matrix elements
    m_InertialtoBody[0][0] =  cb * cc;
    m_InertialtoBody[0][1] =  cb * sc;
    m_InertialtoBody[0][2] = -sb;

    m_InertialtoBody[1][0] =  sa * sb * cc - ca * sc;
    m_InertialtoBody[1][1] =  sa * sb * sc + ca * cc;
    m_InertialtoBody[1][2] =  sa * cb;

    m_InertialtoBody[2][0] =  ca * sb * cc + sa * sc;
    m_InertialtoBody[2][1] =  ca * sb * sc - sa * cc;
    m_InertialtoBody[2][2] =  ca * cb;

    // Initialize the quaternion
    Vector<4>& quat = m_Quat.ic;
    quat[0] = 0.5 * sqrt(1 + m_InertialtoBody[0][0] + m_InertialtoBody[1][1] +
        m_InertialtoBody[2][2]);

    quat[1] = (m_InertialtoBody[0][1] - m_InertialtoBody[1][0]) / (4 * quat[0]);
    quat[2] = (m_InertialtoBody[2][0] - m_InertialtoBody[0][2]) / (4 * quat[0]);
    quat[3] = (m_InertialtoBody[1][2] - m_InertialtoBody[2][1]) / (4 * quat[0]);
}

void RigidBody::Compute(Vector<3> translational_accel, Vector<3> angular_accel)
{
    // Derivatives of states
    m_Pos.der = m_Vel;
    m_Vel.der = m_InertialtoBody.Transpose()*translational_accel;
    m_BodyRate.der = angular_accel;

    // Body rates in rad/sec
    Vector<3>& body_rate = Vector<3>(m_BodyRate);
    double p = body_rate[0];
    double q = body_rate[1];
    double r = body_rate[2];

    // Quaternion derivative
    Vector<4>& quat = Vector<4>(m_Quat);

    double q0q0 = quat[0]*quat[0];
    double q1q1 = quat[1]*quat[1];
    double q2q2 = quat[2]*quat[2];
    double q3q3 = quat[3]*quat[3];

    double e = 1 - q0q0 - q1q1 - q2q2 - q3q3;

    Vector<4>& quat_der = m_Quat.der;
    quat_der[0] = 0.5*(-quat[3]*p - quat[2]*q - quat[1]*r + e*quat[0]);
    quat_der[1] = 0.5*(-quat[2]*p + quat[3]*q + quat[0]*r + e*quat[1]);
    quat_der[2] = 0.5*( quat[1]*p + quat[0]*q - quat[3]*r + e*quat[2]);
    quat_der[3] = 0.5*( quat[0]*p - quat[1]*q + quat[2]*r + e*quat[3]);

    // ECI to Body transformation matrix
    m_InertialtoBody[0][0] = q0q0 - q1q1 - q2q2 + q3q3;
    m_InertialtoBody[0][1] = 2*(quat[0]*quat[1] + quat[2]*quat[3]);
    m_InertialtoBody[0][2] = 2*(quat[1]*quat[3] - quat[0]*quat[2]);

    m_InertialtoBody[1][0] = 2*(quat[2]*quat[3] - quat[0]*quat[1]);
    m_InertialtoBody[1][1] = q0q0 - q1q1 + q2q2 - q3q3;
    m_InertialtoBody[1][2] = 2*(quat[1]*quat[2] + quat[0]*quat[3]);

    m_InertialtoBody[2][0] = 2*(quat[0]*quat[2] + quat[1]*quat[3]);
    m_InertialtoBody[2][1] = 2*(quat[1]*quat[2] - quat[0]*quat[3]);
    m_InertialtoBody[2][2] = q0q0 + q1q1 - q2q2 - q3q3;

    // Euler angles
    m_Euler[0] = atan2( m_InertialtoBody[1][2], m_InertialtoBody[2][2]);
    m_Euler[1] = atan2(-m_InertialtoBody[0][2],
        sqrt(pow(m_InertialtoBody[1][2], 2) + pow(m_InertialtoBody[2][2], 2)));
    m_Euler[2] = atan2( m_InertialtoBody[0][1], m_InertialtoBody[0][0]);
}

Vector<3> RigidBody::GetPos()
{
    Vector<3> result;

    for (int i=0; i<3; i++)
        result[i] = Vector<3>(m_Pos)[i];

    return result;
}

Vector<3> RigidBody::GetVel()
{
    Vector<3> result;

    for (int i=0; i<3; i++)
        result[i] = Vector<3>(m_Vel)[i];

    return result;
}

Vector<3> RigidBody::GetBodyRate()
{
    Vector<3> result;

    for (int i=0; i<3; i++)
        result[i] = Vector<3>(m_BodyRate)[i];

    return result;
}
