// Turntable simulation
// Simulate a turntable driven by a DC motor with a digital controller.
// Jim Ledin                                             December, 2000

#include <dssl.h>

#include <cstdio>
#include <cmath>

double digital_controller(double pos_cmd, double pos_measurement)
{
    static StateList state_list;
    static State<> controller_num(&state_list, State<>::DISCRETE),
        controller_deriv(&state_list, State<>::DISCRETE);

    // Compute the output error
    double error = pos_cmd - pos_measurement;

    // Derivative estimate transfer function. Note that we set the ".der" member
    // of the State variables here even though these are really the "next frame"
    // values for the DISCRETE update method.
    controller_num.der = error;
    controller_deriv.der = 20*(error - controller_num) + 0.9048*controller_deriv;

    // The control law
    double dac_input = 100*error + 50*controller_deriv;

    // Simulate the DAC output
    double dac_output = limit(quantize(dac_input, 20.0/256),
        -10.0, 10.0-20.0/256);

    // Update difference equations
    state_list.Integrate();

    return dac_output;
}

int main()
{
    static StateList state_list;
    static State<> motor_accel(&state_list, State<>::RK4),
        motor_vel(&state_list, State<>::RK4),
        motor_pos(&state_list, State<>::RK4);

    // Constants for frame times and multiframe ratio
    const double step_time = 0.001, controller_step_time = 0.005, end_time = 10.0;
    const int frame_ratio = nint(controller_step_time / step_time);

    // Constants related to model parameters
    const double J = 0.012;         // kg-m^2/sec^2
    const double b = 0.2;           // N-m-sec
    const double K = 0.013;         // N-m/amp
    const double R = 1.1;           // ohms
    const double L = 0.6;           // Henries

    // Compute the transfer function coefficients
    const double a2 = J*L;
    const double a1 = J*R + b*L;
    const double a0 = b*R + K*K;

    // Other constants
    const double pi = 4*atan(1);
    const double pos_cmd = 1;       // radians

    state_list.Initialize(step_time);

    // Open the output file
    FILE* iov = fopen("Turntable.csv", "w");
    if (!iov)
    {
        printf("Unable to open output file 'Turntable.csv'");
        return 1;
    }

    fprintf(iov, "time, dac_output, motor_pos\n");
    int frame_count = 0;
    for(;;)
    {
        // Measure the position of the turntable using the position encoder
        double pos_measurement = quantize(double(motor_pos), 2*pi/4096);

        // Execute the digital controller at the proper time interval
        static double dac_output = 0;
        if (frame_count % frame_ratio == 0)
            dac_output = digital_controller(pos_cmd, pos_measurement);

        // Simulate the 4 millisecond computational delay of the controller
        double delayed_dac_output = transport_delay<double, 4>(dac_output);

        // Write output to file
        fprintf(iov, "%6.3lf, %9.6lf, %9.6lf\n", state_list.Time(),
            delayed_dac_output,  double(motor_pos));

        // Model the motor transfer function
        motor_accel.der = (K*delayed_dac_output - a1*motor_accel -
            a0*motor_vel) / a2;
        motor_vel.der = motor_accel;
        motor_pos.der = motor_vel;

        // Perform end-of-frame activities
        if (state_list.Time() >= end_time)
            break;

        state_list.Integrate();
        frame_count++;
    }

    fclose(iov);
    printf("Output file is Turntable.csv\n");
    return 0;
}